/*
 *     square-, J1-J2-, checkerboard- lattice
 */

#include <stdlib.h>	/* Achtung bei long<->int-Mischmasch */
#include <stdio.h>

#define NN 256		/* maximum number of sites	*/
#define Nw (4*NN)	/* J1-J2: (4*NN),  J1: (2*NN) 	*/

int  pos[NN][3];	/* site-position = (x,y,z)	*/
int  iw[Nw][2];		/* array bonds			*/
int  wj[Nw];		/* nhv faster if wj4=wj/4	*/

#define F0(txt)	{ fprintf(stderr,"\nERROR L%d: "txt"\n",__LINE__);exit(1); }

int e_cell[2][2]={{4,0},{0,4}};

int vertex[NN][2];              /* position of vertizees */         
int nn,nw;
int checker=0;

/* 
+++TeX+++

 Beispiel in zwei Dimensionen. $a=(a_x,a_y)$ und $b=(b_x,b_y)$ sind die 
 zwei Vektoren, die die Elementarzelle aufspannen. 
 Sie sind im allgemeinen nicht senkrecht zueinander. 
 Der Vektor $r=(r_x,r_y)$ liegt innerhalb der Elementarzelle, wenn seine
 Komponenten $(r_a,r_b)$ bezueglich der Elementarvektoren zwischen $0$ und $1$
 liegen.
 \begin{eqnarray*}
      r &=& r_a a   + r_b b \\
 %
    a r &=& r_a a^2 + r_b a b \\
    b r &=& r_a a b + r_b b^2 \\
 %
    r_a &=& \frac{(b^2a-(a,b)b)r}{a^2b^2-(a,b)^2} \\
    r_b &=& \frac{(a^2b-(a,b)a)r}{a^2b^2-(a,b)^2} \\
 \end{eqnarray*}
 Grenzfall $(a,b)=0$ liefert bekannte Beziehungen. 
---TeX--- */

/* vertex -> (x,y) */
int get_koord(const int i,int s[2]){
  if(i<0 || i>=nn) { printf("error: get_koord\n"); return -1; }
  s[0]=vertex[i][0];
  s[1]=vertex[i][1];
  return 0;}
  
/* (x,y,z) -> vertex */  
int get_vertex(int xx[2],int dx,int dy,int dz){
  int i,x[2],(*e)[2]=e_cell,aa,ab,bb,ar,br,ra,rb,det,(*s)[2]=vertex;
  x[0]=xx[0]+dx;x[1]=xx[1]+dy;if(dz!=0)F0("dz!=0");
  aa=e[0][0]*e[0][0]+e[0][1]*e[0][1];
  ab=e[0][0]*e[1][0]+e[0][1]*e[1][1];
  bb=e[1][0]*e[1][0]+e[1][1]*e[1][1];
  det= bb*aa - ab*ab;
  ar=e[0][0]*x[0]   +e[0][1]*x[1] ;
  br=e[1][0]*x[0]   +e[1][1]*x[1] ;
  ra= ( + bb*ar - ab*br );
  rb=-( + ab*ar - aa*br );
  if(ra<0)ra=-((-ra-1)/det)-1;else ra/=det;
  if(rb<0)rb=-((-rb-1)/det)-1;else rb/=det;
  x[0]-=ra*e[0][0]+rb*e[1][0];
  x[1]-=ra*e[0][1]+rb*e[1][1];
  for(i=0;i<nn;i++) 
    if(x[0]==s[i][0] && x[1]==s[i][1]) return i;
#if 0
  printf("error get_vertex: %2d %2d -(%2d %2d)-> %2d %2d not found\n",
    xx[0]+dx,xx[1]+dy,ra,rb,x[0],x[1]);
#endif
  return -1;}

int vertex_init(){ /* vertex_init */
   int i,j,k,r[2],ra,rb,aa,ab,bb,ar,br,det,emin[2],emax[2],
     (*s)[2]=vertex;
   int (*e)[2]=e_cell;
   /* plaetze finden! */
   
   for(i=0;i<2;i++)printf("# %c=(%2d %2d)",'a'+i,e[i][0],e[i][1]);
   i=( e[0][0]*e[1][1]     /*   ax by */ 
      -e[0][1]*e[1][0] );  /* - ay bx */
   printf("#  NN=%d == volume=Det(l)=%d\n",nn, i);
   if(abs(i)!=nn) F0("confusion in cluster size");
   
   /* einhuellenden Kubus ermitteln (alle eck-koords checken) */
   for(i=0;i<2;i++){ 
     emin[i]=emax[i]=e[0][i]+e[1][i]; 
     if(emin[i]>0)emin[i]=0;
     if(emax[i]<0)emax[i]=0;
     for(j=0;j<2;j++){
       if(e[j][i]<emin[i])emin[i]=e[j][i];
       if(e[j][i]>emax[i])emax[i]=e[j][i];
       for(k=j+1;k<2;k++){
         if(e[j][i]+e[k][i]<emin[i])emin[i]=e[j][i]+e[k][i];
         if(e[j][i]+e[k][i]>emax[i])emax[i]=e[j][i]+e[k][i];
   }}}
   printf("# emin=(%d %d) emax=(%d %d)\n",
   emin[0],emin[1],emax[0],emax[1]);
   
   i=0;
   aa=e[0][0]*e[0][0]+e[0][1]*e[0][1];
   ab=e[0][0]*e[1][0]+e[0][1]*e[1][1];
   bb=e[1][0]*e[1][0]+e[1][1]*e[1][1];
   det= bb*aa - ab*ab;

   /* sublattice A */
   for(r[0]=emin[0];r[0]<=emax[0];r[0]+=1)
   for(r[1]=emin[1];r[1]<=emax[1];r[1]+=1){  /* ineffektiv! */
    ar=e[0][0]*r[0]   +e[0][1]*r[1] ;
    br=e[1][0]*r[0]   +e[1][1]*r[1] ;
    ra= ( + bb*ar - ab*br );
    rb=-( + ab*ar - aa*br );
#if 0
    printf("# r=(%2d %2d) det^2(ra rb)=(%4d %4d)",r[0],r[1],ra,rb);
    printf("# ra*a+rb*b=(%2d %2d)", (ra*e[0][0]+rb*e[1][0])/det,
        			   (ra*e[0][1]+rb*e[1][1])/det );
    if(ra>=0 && rb>=0 && ra<det && rb<det) printf("# i=%3d", i);printf("\n");
#endif
    if(ra>=0 && rb>=0 && ra<det && rb<det){
      pos[i][0]=s[i][0]=r[0];
      pos[i][1]=s[i][1]=r[1]; i++; }
   }
   
   if(i!=nn)fprintf(stderr,"error vertex_init i!=N i=%d N=%d\n",i,nn);

   return 0;
}


  /*  i=site 0..NN, d=direction 0..2, s=steps ..,-1,0,1,.. rc=i+s*vec_d */
  /* besser site(i,dx,dy[,dz])  dy<0 geht nicht !!!!  */
  int site_move(int i,int dx,int dy,int dz,...){ int r[2];
    if(i<0 || i>=nn)return -1;
    get_koord(i,r);
    i=get_vertex(r,dx,dy,dz);
    return(i);}

  /* Abstandsdefinition (z.B.) fuer Untergittermagnetisierung */
  int site_dist(int i,int j){ /* Abstandsdefinition: |dx|+|dy|+|dz| */
     int ri[2],rj[2];
     get_koord(i,ri);
     get_koord(j,rj);
     return(abs(ri[0]-rj[0])+abs(ri[1]-rj[1]));}

  int initvars(){
    int i; 
    vertex_init();
#define Bd(l,a,b) for(i=0;i<nn;i++){ iw[l][0]=a; iw[l][1]=b; }
    Bd(i     ,i,site_move(i,1, 0, 0));
    Bd(i+1*nn,i,site_move(i,0, 1, 0));
    if( nw>2*nn ){
      Bd(i+2*nn,i,site_move(i,1, 1, 0));
      Bd(i+3*nn,i,site_move(i,1,-1, 0));
    }
    for(i=0;i< nw;i++)
      wj[i]=(i<2*nn)?1:((checker && ((vertex[iw[i][0]][0]
                                     +vertex[iw[i][0]][1])&1)==(i/nn&1))?3:2);
    return 0;}    

/* 
  ==================================================================
*/

int outmodel(){
  int i;
  /* Ausgabe Modell in Maschinen-lesbarer Form (fuer allg. Model.H) */
  printf("#===uvec=== 1 0 0 1\n");
  printf("#===edge=== %d %d %d %d\n",e_cell[0][0],e_cell[0][1],e_cell[1][0],e_cell[1][1]);
  printf("#===pbcf=== 0 0 %d %d %d %d %d %d\n",e_cell[0][0],e_cell[0][1],e_cell[0][0]+e_cell[1][0],e_cell[0][1]+e_cell[1][1],e_cell[1][0],e_cell[1][1]);
  printf("#+++model+++\n");
  printf("# NN=Number_of_sites Nw=Number_of_bonds Nv=Number_of_bonds(2)\n");
  printf("%2u %3u 0\n",nn,nw);
  printf("# s  X  Y  Z p_e p_U p_f (s=site  (X Y Z)=position  p=parameter-Nr)\n");
  for(i=0;i<nn;i++)
  printf(" %2u %2g %2g %2d 3 4 5\n",i,pos[i][0]+.5,pos[i][1]+.5,pos[i][2]);
  printf("# Bond s1 s2 p_t p_J p_V\n");
  for(i=0;i<nw;i++)
  printf(" %3u %2d %2d  2 %2d  %c0\n",i,iw[i][0],iw[i][1],wj[i],
   (  (abs(pos[iw[i][0]][0]-pos[iw[i][1]][0])<=1
    && abs(pos[iw[i][0]][1]-pos[iw[i][1]][1])<=1 )?' ':'+'));
  printf("#---model---\n");
  return(0); }

int get_det(int e[2][2]){
  return e[0][0]*e[1][1]-e[0][1]*e[1][0]; 
}

int main(int argn,char *argv[]){
  int i;
  printf("# m_square v0.2 Feb2001 JS\n");
  if(argn!=6){
    fprintf(stderr,
    "usage:   m_square l11 l12 l21 l22 bounds\n"
    "example: m_square 4 0  0 4  2\n"
    "   edgevector=((l11,l12),(l21,l22))\n"
    "   bounds: 1=J1 2=J1+J2 3=checkerboard\n"
    );
    return 1;
  }
  for(i=0;i<4;i++)e_cell[i/2][i%2]=atoi(argv[i+1]);
  nn=get_det(e_cell);
  if(atoi(argv[5])==3) checker=1;	/* checkerboard */
  if(atoi(argv[5])>=2) nw=4*nn;		/* J1+J2, checkerboard */
  else                 nw=2*nn;		/* J1 */
  initvars();
  outmodel();
  return 0;
}
