#ifndef Model_c
#define NN       16	/* 4d0 8d2 10h7 16q0 18t3 20d4 26h5 32q4 34h12 36s0 */
#define N1        4	/* Betts_2 cyclenzahl (chain=1) */
#define N2        0	/* Betts_3 offset     (chain=0) */
#define Title     "JJ-16q0-Betts-Model"    /* 123456789=hdtqps..o. */
#define Sym       4	/* x x 7 x x x x x x Symmetrien */
#define Sud       1	/* 1=ud-symmetry 0=no */
#define CM(JJ,tJ,tU) JJ	  /* Heisenberg-,  t-J- or  Hubbard-model */

#if N1>1 || N2>0
#define Nw (6*NN)	/* (6*NN)=J1+J2+J3  (4*NN)=J1+J2  (2*NN)=only J1 */
#elif NN>3
#define Nw (2*NN)	/* (2*NN)=J1-J2-chain	(NN)=J1-chain  */
#else
#define Nw NN		/* NN=J1-chain */
#endif
#define Ne NN
#define SBase 1		/* SBase/noSBase = with/without symmetry */ 
#endif
  int site_move(int i,int dx,int dy,...);	/* declaration */
#ifdef Model_c
/* ================== 2D-Gitter nach Betts =========================== */
/* => Betts-Notation example: 14h4 => N=14,N1=1,N2=4 => 1 helix, dy=4 */
  /*  i=site 0..NN, d=direction 0..2, s=steps ..,-1,0,1,.. rc=i+s*vec_d */
  /* besser site(i,dx,dy[,dz])  dy<0 geht nicht !!!!  */
  int site_move(int i,int dx,int dy,...){ 
    if(i<0 || i>=NN)return -1;while(dx<0)dx+=NN/N1;while(dy<0)dy+=N1;
    i=(i+dx)%(NN/N1)+(i/(NN/N1))*(NN/N1);		/* x-direction */
    i=(i+dy+((i/(NN/N1)+dy)/N1)*(N2-N1+NN/N1))%(NN/N1)+
                   ((i/(NN/N1)+dy)%N1)*(NN/N1);		/* y-direction */
    return(i);}

  int initvars(){
    int i,j; 
#if NN==16 && N1==4 && N2==0
    static int tis[8][NN]={
     {  2, 3, 0, 1, 6, 7, 4, 5,10,11, 8, 9,14,15,12,13},  /* Tx^2 */
     { 10,11, 8, 9,14,15,12,13, 2, 3, 0, 1, 6, 7, 4, 5},  /* Ty^2 */
     {  1, 0, 3, 2, 7, 6, 5, 4, 9, 8,11,10,15,14,13,12},  /* Tx*vertrefl */
     {  5, 6, 7, 4, 3, 0, 1, 2,13,14,15,12,11, 8, 9,10},  /* Ty*vertrefl */
     {  0, 3, 2, 1, 6, 5, 4, 7, 8,11,10, 9,14,13,12,15},  /* vert. refl */
     {  0, 1, 2, 3,14,15,12,13, 8, 9,10,11, 6, 7, 4, 5},  /* hori. refl */
     {  0, 5,10,15,12, 1, 6,11, 8,13, 2, 7, 4, 9,14, 3},  /* diag  refl */
     {  0, 5,10,15,14, 3, 4, 9, 8,13, 2, 7, 6,11,12, 1}};  /* dreh links */
    for(i=0;i<7 && i<Sym;i++)for(j=0;j<NN;j++)sym.i[i][j]=tis[i][j];
#elif NN==28 && N1==2 && N2==8
    static int tis[Sym][NN]={
 {7,8,9,10,11,12,13,0,1,2,3,4,5,6,21,22,23,24,25,26,27,14,15,16,17,18,19,20},
 {18,19,20,21,22,23,24,25,26,27,14,15,16,17,10,11,12,13,0,1,2,3,4,5,6,7,8,9},
 {1,0,13,12,11,10,9,8,7,6,5,4,3,2,23,22,21,20,19,18,17,16,15,14,27,26,25,24}};
    for(i=0;i<Sym;i++)for(j=0;j<NN;j++)sym.i[i][j]=tis[i][j];
#endif    
    nu=NN/2;nd=NN/2+NN%2;
#define Bd(l,a,b) for(i=0;i<NN;i++){iw[l][0]=a;iw[l][1]=b;}
#if N1>1 || N2>0
    Bd(i     ,i,site_move(i,1, 0));
    Bd(i+  NN,i,site_move(i,0, 1));
    Bd(i+2*NN,i,site_move(i,1, 1));
    Bd(i+3*NN,i,site_move(i,-1,1));
#if Nw>4*NN
    Bd(i+4*NN,i,site_move(i,2,0));
    Bd(i+5*NN,i,site_move(i,0,2));
#endif    
#else	/* Kette */
#if 0
    if(Sym>0)
      for(j=0;j<NN;j++)sym.i[0  ][j]=(j+NN/2)%NN;	/* Tx^(NN/2) */
    for(i=0;i<Sym-1;i++)
      for(j=0;j<NN;j++)sym.i[1+i][j]=(NN+2*i-j+1)%NN;	/* reflektions */
#endif
    Bd(i   ,i,site_move(i,1, 0));
#if Nw>NN    
    Bd(i+NN,i,site_move(i,2, 0));
#endif
#endif
#if N1>1 || N2>0
    for(i=0;i< Nw;i++)wj[i]=(i<2*NN)? 1:(i<4*NN)?2 : 3;
    for(i=0;i< Nw;i++)wt[i]=(i<2*NN)? 4:(i<4*NN)?5 : 6;
    for(i=0;i< NN;i++){ pos[i][0]=(i-i/(NN/N1))%(NN/N1);
                        pos[i][1]=i/(NN/N1);pos[i][2]=0; }
#else
    for(i=0;i< Nw;i++)wj[i]=(i<NN)? 1:(i<2*NN)?2 : 3;
    for(i=0;i< Nw;i++)wt[i]=(i<NN)? 4:(i<2*NN)?5 : 6;
    for(i=0;i< NN;i++){ pos[i][0]=i;pos[i][1]=pos[i][2]=0; }
#endif
    for(i=0;i< NN;i++)wu[i]= 7;
    for(j=0;j< NN;j++)we[j]= 8;
    numpara=CM(4,7,9);
    para[0]= 0.0;	/* immer 0 */
    para[1]= 1.0;	/* J1 */
    para[2]= 0.4;	/* J2 */
    para[3]= 0.0;	/* J3 */
    para[4]=-1.0;	/* t1 */
    para[5]= 0.15;	/* t2 */
    para[6]=-0.2;	/* t3 */
    para[7]= 8.0;	/* e */
    para[8]= 0.0;	/* U */
    return 0;}    
#endif
